from onegov.core.custom import json
from onegov.user.auth import Auth
from onegov.user.collections import UserCollection
from onegov.user.forms import LoginForm
from onegov.user.forms import PasswordResetForm
from onegov.user.forms import RegistrationForm
from onegov.user.forms import RequestPasswordResetForm
from onegov.user.forms import UserGroupForm
from onegov.user.models import User
from onegov.user.models import UserGroup


class DummyApp():
    def __init__(self, session):
        self._session = session

    def session(self):
        return self._session


class DummyRequest():
    def __init__(self, session):
        self.app = DummyApp(session)
        self.client_addr = '127.0.0.1'

    def load_url_safe_token(self, token, **kwargs):
        if not token:
            return None

        return json.loads(token)


class DummyPostData(dict):
    def getlist(self, key):
        v = self[key]
        if not isinstance(v, (list, tuple)):
            v = [v]
        return v


def test_login_form():
    # Test validation
    form = LoginForm()
    assert not form.validate()

    form.process(DummyPostData({
        'username': 'info@example.com',
        'password': 'much_secret'
    }))
    assert form.validate()
    assert form.login_data == {
        'username': 'info@example.com',
        'password': 'much_secret',
        'second_factor': None,
    }

    form.process(DummyPostData({
        'username': 'info@example.com',
        'password': 'much_secret',
        'yubikey': 'abcdefghijklmnopqrstuvwxyz'
    }))
    assert form.validate()
    assert form.login_data == {
        'username': 'info@example.com',
        'password': 'much_secret',
        'second_factor': 'abcdefghijklmnopqrstuvwxyz',
    }


def test_registration_form(session):
    # Test validation
    form = RegistrationForm()
    assert not form.validate()

    form.process(DummyPostData({
        'username': 'info@example.com',
        'password': 'much_secret',
        'confirm': 'very_secret'
    }))
    assert not form.validate()

    form.process(DummyPostData({
        'username': 'info@example.com',
        'password': 'much_secret',
        'confirm': 'much_secret',
        'roboter_falle': 'fooled'
    }))
    assert not form.validate()

    form.process(DummyPostData({
        'username': 'info@example.com',
        'password': 'much_secret',
        'confirm': 'much_secret'
    }))
    assert form.validate()

    # Test register user
    auth = Auth(session, 'foobar')
    auth.register(form, DummyRequest(session))

    assert session.query(User).filter_by(username='info@example.com').one()


def test_request_password_reset_form():
    # Test validation
    form = RequestPasswordResetForm()
    assert not form.validate()

    form.process(DummyPostData({
        'email': 'name',
    }))
    assert not form.validate()

    form.process(DummyPostData({
        'email': 'info@example.com',
    }))
    assert form.validate()


def test_password_reset_form(session):
    # Test validation
    request = DummyRequest(session)
    form = PasswordResetForm()
    assert not form.validate()

    form.process(DummyPostData({
        'email': 'name',
        'password': 'secret',
    }))
    assert not form.validate()

    form.process(DummyPostData({
        'email': 'info@example.com',
        'password': 'secret',
    }))
    assert not form.validate()

    form.process(DummyPostData({
        'email': 'info@example.com',
        'password': 'much_secret',
    }))
    assert form.validate()
    assert not form.update_password(request)

    # Test update password
    form.process(DummyPostData({
        'email': 'info@example.com',
        'password': 'much_secret',
        'token': json.dumps({'username': 'username'})
    }))
    assert form.validate()
    assert not form.update_password(request)

    form.process(DummyPostData({
        'email': 'info@example.com',
        'password': 'much_secret',
        'token': json.dumps({'username': 'info@example.com'})
    }))
    assert form.validate()
    assert not form.update_password(request)

    form.process(DummyPostData({
        'email': 'info@example.com',
        'password': 'much_secret',
        'token': json.dumps({'username': 'info@example.com'})
    }))
    assert form.validate()
    assert not form.update_password(request)

    assert UserCollection(session).register(
        'info@example.com', 'very_secret', request
    )
    form.process(DummyPostData({
        'email': 'info@example.com',
        'password': 'much_secret',
        'token': json.dumps({'username': 'info@example.com'})
    }))
    assert form.validate()
    assert not form.update_password(request)

    form.process(DummyPostData({
        'email': 'info@example.com',
        'password': 'much_secret',
        'token': json.dumps({
            'username': 'info@example.com',
            'modified': 'now'
        })
    }))
    assert form.validate()
    assert not form.update_password(request)

    form.process(DummyPostData({
        'email': 'info@example.com',
        'password': 'much_secret',
        'token': json.dumps({
            'username': 'info@example.com',
            'modified': ''
        })
    }))
    assert form.validate()
    assert form.update_password(request)

    form.process(DummyPostData({
        'email': 'info@example.com',
        'password': 'much_secret',
        'token': json.dumps({
            'username': 'info@example.com',
            'modified': ''
        })
    }))
    assert form.validate()
    assert not form.update_password(request)


def test_user_group_form():
    # Test apply / update
    form = UserGroupForm()
    group = UserGroup(name='Group X')

    form.apply_model(group)
    assert form.name.data == 'Group X'

    form.name.data = 'Group Y'
    form.update_model(group)
    assert group.name == 'Group Y'

    # Test validation
    form = UserGroupForm()
    assert not form.validate()

    form = UserGroupForm(DummyPostData({'name': 'Group'}))
    assert form.validate()



from onegov.user.models import UserGroup
from onegov.core.collection import GenericCollection


class UserGroupCollection(GenericCollection):
    """ Manages a list of user groups.

    Use it like this::

        from onegov.user import UserGroupCollection
        groups = UserGroupCollection(session)

    """

    @property
    def model_class(self):
        return UserGroup

    def query(self):
        query = super(UserGroupCollection, self).query()
        return query.order_by(UserGroup.name)



import morepath

from datetime import datetime
from itsdangerous import URLSafeSerializer, BadSignature
from onegov.core.utils import relative_url
from onegov.user import log
from onegov.user.collections import UserCollection
from onegov.user.errors import ExpiredSignupLinkError
from onegov.user.utils import is_valid_yubikey


class Auth(object):
    """ Defines a model for authentication methods like login/logout.
    Applications should use this model to implement authentication views.

    """

    identity_class = morepath.Identity

    def __init__(self, session, application_id, to='/', skip=False,
                 yubikey_client_id=None, yubikey_secret_key=None,
                 signup_token=None, signup_token_secret=None):
        assert application_id  # may not be empty!

        self.session = session
        self.application_id = application_id

        self.signup_token = signup_token
        self.signup_token_secret = signup_token_secret

        # never redirect to an external page, this might potentially be used
        # to trick the user into thinking he's on our page after entering his
        # password and being redirected to a phising site.
        self.to = relative_url(to)
        self.skip = skip

        self.yubikey_client_id = yubikey_client_id
        self.yubikey_secret_key = yubikey_secret_key

    @classmethod
    def from_app(cls, app, to='/', skip=False, signup_token=None):
        return cls(
            session=app.session(),
            application_id=app.application_id,
            yubikey_client_id=getattr(app, 'yubikey_client_id', None),
            yubikey_secret_key=getattr(app, 'yubikey_secret_key', None),
            to=to,
            skip=skip,
            signup_token=signup_token,
            signup_token_secret=getattr(app, 'identity_secret', None)
        )

    @classmethod
    def from_request(cls, request, to='/', skip=False, signup_token=None):
        return cls.from_app(request.app, to, skip, signup_token)

    @classmethod
    def from_request_path(cls, request, skip=False, signup_token=None):
        return cls.from_request(
            request, request.transform(request.path), skip, signup_token)

    @property
    def signup_token_serializer(self):
        assert self.signup_token_secret
        return URLSafeSerializer(self.signup_token_secret, salt='signup')

    def new_signup_token(self, role, max_age=24 * 60 * 60, max_uses=1):
        """ Returns a signup token which can be used for users to register
        themselves, directly gaining the given role.

        Signup tokens are recorded on the user to make sure that only the
        requested amount of uses is allowed.

        """
        return self.signup_token_serializer.dumps({
            'role': role,
            'max_uses': max_uses,
            'expires': int(datetime.utcnow().timestamp()) + max_age
        })

    def decode_signup_token(self, token):
        try:
            return self.signup_token_serializer.loads(token)
        except BadSignature:
            return None

    @property
    def users(self):
        return UserCollection(self.session)

    def redirect(self, request):
        return morepath.redirect(request.transform(self.to))

    def skippable(self, request):

        if not self.skip:
            return False

        # this is the default paramter, we won't skip to it in any case
        if self.to == '/':
            return False

        try:
            return request.has_access_to_url(self.to)
        except KeyError:
            return False

    def is_valid_second_factor(self, user, second_factor_value):
        if not user.second_factor:
            return True

        if not second_factor_value:
            return False

        if user.second_factor['type'] == 'yubikey':
            return is_valid_yubikey(
                client_id=self.yubikey_client_id,
                secret_key=self.yubikey_secret_key,
                expected_yubikey_id=user.second_factor['data'],
                yubikey=second_factor_value
            )
        else:
            raise NotImplementedError

    def login(self, username, password, client='unknown', second_factor=None):
        """ Takes the given username and password and matches them against
        the users collection.

        :meth:`login_to` should be the preferred method in most cases, as it
        sets up automatic logging of login attempts and provides optional
        login/logout messages.

        :return: An identity bound to the ``application_id`` if the username
        and password match.

        """

        user = self.users.by_username_and_password(username, password)

        def fail():
            log.info("Failed login by {} ({})".format(client, username))
            return None

        if user is None:
            return fail()

        if not user.active:
            return fail()

        if not self.is_valid_second_factor(user, second_factor):
            return fail()

        log.info("Successful login by {} ({})".format(client, username))

        return self.identity_class(
            userid=user.username,
            role=user.role,
            application_id=self.application_id
        )

    def login_to(self, username, password, request, second_factor=None):
        """ Matches the username and password against the users collection,
        just like :meth:`login`. Unlike said method it returns no identity
        however, instead a redirect response is returned which will remember
        the identity.

        :return: A redirect response to ``self.to`` with the identity
        remembered as a cookie. If not successful, None is returned.

        """

        identity = self.login(username, password, request.client_addr,
                              second_factor)

        if identity is None:
            return None

        response = self.redirect(request)
        request.app.remember_identity(response, request, identity)

        user = self.users.by_username(username)
        if user:
            user.save_current_session(request)

        return response

    def logout_to(self, request):
        """ Logs the current user out and redirects to ``self.to``.

        :return: A response redirecting to ``self.to`` with the identity
        forgotten.

        """

        user = self.users.by_username(request.identity.userid)
        if user:
            user.remove_current_session(request)

        response = self.redirect(request)
        request.app.forget_identity(response, request)

        return response

    @property
    def permitted_role_for_registration(self):
        if not self.signup_token:
            return 'member'

        assert self.signup_token_secret
        params = self.decode_signup_token(self.signup_token)

        if not params:
            return None

        if params['expires'] < int(datetime.utcnow().timestamp()):
            return None

        signups = UserCollection(self.session)\
            .by_signup_token(self.signup_token).count()

        if signups >= params['max_uses']:
            return None

        return params['role']

    def register(self, form, request):
        """ Registers the user using the information on the registration form.

        Takes the signup token into account to provide the user with the
        proper role.

        See :meth:`onegov.user.collections.UserCollection.register_user` for
        more information.

        """

        role = self.permitted_role_for_registration

        if role is None:
            raise ExpiredSignupLinkError()

        return UserCollection(self.session).register(
            username=form.username.data,
            password=form.password.data,
            request=request,
            role=role,
            signup_token=self.signup_token
        )



from datetime import datetime
from onegov.core.crypto import hash_password, verify_password
from onegov.core.orm import Base
from onegov.core.orm.mixins import data_property, TimestampMixin
from onegov.core.orm.types import JSON, UUID, LowercaseText
from onegov.core.security import forget, remembered
from onegov.core.utils import remove_repeated_spaces
from onegov.search import ORMSearchable
from onegov.user.models.group import UserGroup
from onegov.user.utils import is_valid_yubikey_format
from onegov.user.utils import yubikey_otp_to_serial
from sqlalchemy import Boolean, Column, Index, Text, func, ForeignKey
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import backref, deferred, relationship
from uuid import uuid4


class User(Base, TimestampMixin, ORMSearchable):
    """ Defines a generic user. """

    __tablename__ = 'users'

    #: the type of the item, this can be used to create custom polymorphic
    #: subclasses of this class. See
    #: `<http://docs.sqlalchemy.org/en/improve_toc/\
    #: orm/extensions/declarative/inheritance.html>`_.
    type = Column(Text, nullable=True)

    __mapper_args__ = {
        'polymorphic_on': type
    }

    es_properties = {
        'username': {'type': 'text'},
        'realname': {'type': 'text'},
        'userprofile': {'type': 'text'}
    }
    es_public = False

    @property
    def es_suggestion(self):
        return (self.realname or self.username, self.username)

    @property
    def userprofile(self):
        data = []

        if self.data:
            for value in self.data.values():
                if value and isinstance(value, str):
                    data.append(value)

        return data

    #: the user id is a uuid because that's more secure (no id guessing)
    id = Column(UUID, nullable=False, primary_key=True, default=uuid4)

    #: the username may be any string, but will usually be an email address
    username = Column(LowercaseText, unique=True, nullable=False)

    #: the password is stored with the hashing algorithm defined by onegov.core
    password_hash = Column(Text, nullable=False)

    #: the role is relevant for security in onegov.core
    role = Column(Text, nullable=False)

    #: the group this user belongs to
    group_id = Column(UUID, ForeignKey(UserGroup.id), nullable=True)
    group = relationship(
        UserGroup, backref=backref('users', lazy='dynamic')
    )

    #: the real name of the user for display (use the :attr:`name` property
    #: to automatically get the name or the username)
    realname = Column(Text, nullable=True)

    #: extra data that may be stored with this user, the format and content
    #: of this data is defined by the consumer of onegov.user
    #: by default, this data is only loaded by request, so if you need to
    #: load a lot of data columns, do something like this::
    #:
    #:     session.query(User).options(undefer("data"))
    #:
    data = deferred(Column(JSON, nullable=True))

    #: two-factor authentication schemes are enabled with this property
    #: if no two-factor auth is used, the value is NULL, if one *is* used,
    #: there should be a dictionary with the type of the two-factor
    #: authentication as well as custom values required by the two-factor
    #: implementation.
    #:
    #: e.g.::
    #:
    #:      {
    #:          'type': 'yubikey',
    #:          'data': 'ccccccbcgujh'
    #:      }
    #:
    #: Note that 'data' could also be a nested dictionary!
    #:
    second_factor = Column(JSON, nullable=True)

    #: true if the user is active
    active = Column(Boolean, nullable=False, default=True)

    #: the signup token used by the user
    signup_token = Column(Text, nullable=True, default=None)

    __table_args__ = (
        Index('lowercase_username', func.lower(username), unique=True),
    )

    @hybrid_property
    def title(self):
        """ Returns the realname or the username of the user, depending on
        what's available first. """
        return self.realname or self.username

    @title.expression
    def title(cls):
        return func.coalesce(func.nullif(cls.realname, ''), cls.username)

    @hybrid_property
    def password(self):
        """ An alias for :attr:`password_hash`. """
        return self.password_hash

    @password.setter
    def password(self, value):
        """ When set, the given password in cleartext is hashed using
        onegov.core's default hashing algorithm.

        """
        self.password_hash = hash_password(value)

    def is_matching_password(self, password):
        """ Returns True if the given password (cleartext) matches the
        stored password hash.

        """
        return verify_password(password, self.password_hash)

    @classmethod
    def get_initials(cls, username, realname=None):
        """ Takes the name and returns initials which are at most two
        characters wide.

        Examples:

        admin => A
        nathan.drake@example.org => ND
        Victor Sullivan => VS
        Charles Montgomery Burns => CB

        """

        # for e-mail addresses assume the dot splits the name and use
        # the first two parts of said split (probably won't have a middle
        # name in the e-mail address)
        if not realname:
            username = username.split('@')[0]
            parts = username.split('.')[:2]

        # for real names split by space and assume that with more than one
        # part that the first and last part are the most important to get rid
        # of middlenames
        else:
            parts = remove_repeated_spaces(realname).split(' ')

            if len(parts) > 2:
                parts = (parts[0], parts[-1])

        return ''.join(p[0] for p in parts).upper()

    @property
    def initials(self):
        return self.get_initials(self.username, self.realname)

    @property
    def has_yubikey(self):
        if not self.second_factor:
            return False

        return self.second_factor.get('type') == 'yubikey'

    @property
    def yubikey(self):
        if not self.has_yubikey:
            return None

        return self.second_factor.get('data')

    @yubikey.setter
    def yubikey(self, yubikey):
        if not yubikey:
            self.second_factor = None
        else:
            assert is_valid_yubikey_format(yubikey)

            self.second_factor = {
                'type': 'yubikey',
                'data': yubikey[:12]
            }

    @property
    def yubikey_serial(self):
        """ Returns the yubikey serial of the yubikey associated with this
        user (if any).

        """

        yubikey = self.yubikey

        return yubikey and yubikey_otp_to_serial(yubikey) or None

    #: sessions of this user
    sessions = data_property()

    #: tags of this user
    tags = data_property()

    def cleanup_sessions(self, request):
        """ Removes stored sessions not valid anymore. """

        self.sessions = self.sessions or {}
        for session_id in list(self.sessions.keys()):
            if not remembered(request.app, session_id):
                del self.sessions[session_id]

    def save_current_session(self, request):
        """ Stores the current browser session. """

        self.sessions = self.sessions or {}
        self.sessions[request.browser_session._token] = {
            'address': request.client_addr,
            'timestamp': datetime.utcnow().isoformat(),
            'agent': request.user_agent
        }

        self.cleanup_sessions(request)

    def remove_current_session(self, request):
        """ Removes the current browser session. """

        token = request.browser_session._token
        if self.sessions and token and token in self.sessions:
            del self.sessions[token]

        self.cleanup_sessions(request)

    def logout_all_sessions(self, request):
        """ Terminates all open browser sessions. """

        self.sessions = self.sessions or {}
        for session_id in self.sessions:
            forget(request.app, session_id)

        self.cleanup_sessions(request)



from onegov.form import Form
from onegov.user import _
from wtforms import RadioField, validators
from wtforms.fields.html5 import IntegerField


class SignupLinkForm(Form):
    """ A form to generate signup links for specific roles. """

    role = RadioField(
        label=_("Role"),
        validators=[validators.InputRequired()],
        choices=[
            ('member', ("Member")),
            ('editor', _("Editor")),
            ('admin', _("Admin"))
        ]
    )

    max_age = RadioField(
        label=_("Expires in"),
        validators=[validators.InputRequired()],
        choices=[
            ('hour', _("1 hour")),
            ('day', _("24 hours")),
            ('week', _("7 days")),
            ('month', _("30 days"))
        ]
    )

    max_uses = IntegerField(
        label=_("Number of Signups"),
        validators=[
            validators.InputRequired(),
            validators.NumberRange(1, 10000)
        ],
    )

    def signup_token(self, auth):
        assert self.role.data in ('member', 'editor', 'admin')

        max_age = {
            'hour': 60 * 60,
            'day': 60 * 60 * 24,
            'week': 60 * 60 * 24 * 7,
            'month': 60 * 60 * 24 * 30
        }.get(self.max_age.data, 60 * 60)

        max_uses = int(self.max_uses.data)

        return auth.new_signup_token(self.role.data, max_age, max_uses)
